


// FluxNetz TAN-Server / Shop oder Internet-eMail Frontend Manager


/* Copyright 2003 Uli Funcke & Klaus Kurzmann, FluxNetz GmbH */

/*
 * This file is part of Tanne
 *
 * Tanne is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 *
 * Tanne is distributed in the hope that it will be usefull,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tanne; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <termios.h>
#include <crypt.h>

#include <time.h>
#include <errno.h>
#include <syslog.h>
#include <sys/time.h>

#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>

#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>

#include <sys/wait.h>
#include <sys/stat.h>
#include <signal.h>

#include "config.h"


//____________________________________________________________________________________________________________
// Defines



#define JA 1
#define NEIN 0


#include "version.h"

// define DEBUG if you want some lgging info
// DEBUG and DEBUG2 for more
// and DEBUG3 for massive logging info 
// #define DEBUG
// #define DEBUG2
// #define DEBUG3

// define DEBUG_IN_FILE if you want to log to a file instead of
// syslog / give the name of the file
// #define DEBUG_IN_FILE "/var/log/tanne.log"


// do not use pipe() - use socketpair() instead
#define USE_SOCKETPAIR


// amount of shared memory reserved for tanne ID's
// maximum number of ID's
#define ANZ_IDS 10000


// maximum possible simultanous sessions
#define ANZ_SESSIONS 500


// default umask
#define NORMALE_UMASK 0027


// uid-type. It's system dependent. (is nobody as -2 or as 65534 in /etc/passwd)
#define UID_TYP short


// Tanne-Port = TCP-Port
#define TANNE_PORT 14002


// size in bytes of the ID's
#define ID_LAENGE 32


// maximum bytes for one command
#define MAX_BEFEHL 32


// maximum size of total pathname of tanne backends
#define MAX_ORDNER 80


// size of buffer for input from frontend to tanne
// (must be longer than ID_LAENGE + MAX_BEFEHL + 2 (for '\n' + ASCIIZ))
#define MAX_ZEILE 1024


// define, if read() should get only one byte at once from the socket
// #define EIN_BYTE_LESEN


// else define how much bytes are read at once
#define MAX_LESEBEREICH 2048


// this command was sent to the tanne backend if it
// shuold terminate because tanne has counted down into timeout
#define ENDE_BEFEHL "timeout\n"


// lines which begin with this character mark tanne commands.
// a line containing two SPZ end the session
#define SPZ '|'


// lines which begin with SPZ and this character mark tanne commands which do not
// cancel the other TID's of the session
#define SPZ_AGAIN '!'


// this command is depreached. It marks the end of the session
#define SPZ_CLIENDE '@'


// fast bynary transfer from backend to frontend
#define SPZ_TURBO1 '>'


// fast bynary transfer from frontend to backend
#define SPZ_TURBO2 '<'


// adjust the timeout
#define SPZ_TIMEOUT '$'


// will be sent to frontend if there was an error
#define F_FEHLER "|F|"


// default timeout in 0.5 sec
#ifdef DEBUG
// 30min
#define SESSION_TIME 60
#else
// 15min
#define SESSION_TIME 30
#endif


// maximum timeout (4 hours)
#define SESSION_TIME_MAX 480


// minimum timeout (1 minute)
#define SESSION_TIME_MIN 2



//_____________________________________________________________________________________________________
// nothing to configure below this line


// session status
#define FREI 0
#define BELEGT 1
#define STANDBY 2
#define AKTIV 3
#define ZOMBIE 4
#define AUFRUF_TRIGGER 5



// Befehle-Flag's (Bits in SBefehl[].flag angeben: 1, 2, 4, 8 u.s.w.)

// Multi-Befehl: kein Befehl der Session wird ungltig beim Aufruf des Links
#define BF_MULTI  1

// die Funktion der_select_test() sagt, es gibt Daten...
// ... vom Programm zum Socket
#define SELECT_P2S 1

// ... vom Socket zum Programm
#define SELECT_S2P 2


//_____________________________________________________________________________________________________

// Macros
#define FATAL( x ) { logger( x ); exit( 1 ); }
#define BETRAG( x ) (((x) > 0 ) ? (x) : -(x))




//_____________________________________________________________________________________________________
// Strukturen


// Globale Datenstruktur im shm-Bereich
// Enthlt eine Liste aller gerade gltigen Befehle



typedef struct
{
	signed short sess_id; // Session-id
	short flag;
	char id[ID_LAENGE + 1];
	char bef[MAX_BEFEHL + 1];
}
SBefehl;




// Die Session-id ist der Index des verwendeten Elements aus Session[]
typedef struct
{
	unsigned short status; // 0 = frei, 1 = belegt, 2 = aktiv, 3 = zombie (hand[] mu freigegeben werden)
	long pid; // Prozess-ID des Aufruf-Programms
	long a_pid; // Prozess-ID des clients wenn status == AKTIV
	short ablauf; // Timeout
	short timeout; // default-Timeout
#ifdef USE_SOCKETPAIR
	int hand[2]; // eine socketpair() Verbindung zwischen dem Tanne-Client und dem Programm
#else
	int hs[2]; // pipe() Handle von Tanne zum Programm (Schreiben)
	int hl[2]; // pipe() Handle vom Programm zur Tanne zurck (Lesen)
#endif
}
SSession;





//___________________________________________________________________________________________________
// globale Variablen

extern SBefehl *Befehl;
extern SSession *Session;

extern int shm_id;							// Globale Shared-Memmory-ID
extern int sem_id;							// Globale Semaphore
extern int Gsocke;							// das Socket

#ifdef USE_SOCKETPAIR
extern int GHand;								// Globales Handle der Client-Ende
#else
extern int GHandL;							// Globales Handle der Pipe, Lesen-Ende
extern int GHandS;							// Globales Handle der Pipe, Schreiben-Ende
#endif
extern signed short sess_id;		// interne Session ID
extern struct sockaddr_in in_name;
extern char zeile[];

#ifndef EIN_BYTE_LESEN
extern char lesebereich[2][MAX_LESEBEREICH];
extern short lesezeiger[2];
extern short anz_gelesen[2];
#endif
extern sigset_t signale, sig_alt;
extern long alt_freispeicher;
extern struct sigaction sigh_term;
extern struct sigaction sigh_child;
extern struct sigaction sigh_alarm;
extern struct sigaction sigh_pipe;

extern short sem_in_signal_flag;


//___________________________________________________________________________________________________
// Funktionen



//_________
// ipc_fn.c

int shmStart(void);							// Shared-Memmory erstellen
void shmEnde(void);							// ... freigeben
int shmAn(void);								// Child attached
void shmAus(void);							// detached


// Semaphoren-Operatoren
void P(void);										// Sperren - Warten
void V(void);										// Freigeben



//_________
// netzio.c

void logger(char *str, ...);
void ende(short sessionstatus);

int netz_init(void);
void socken_init(long so);

short lesen(long hand);
void senden(char *txt, long hand);
long bytes_lesen(long hand, long anz);
void bytes_senden(long ol, long hand);

#ifndef EIN_BYTE_LESEN
short zeichenlesen(short i, char *z);
#endif



//__________
// signale.c


void klingel(int na);
void terminator_server(int na);
void terminator_client(int na);
void kind_ende(int na);
void server_signale(void);
void client_signale(void);
void signalgeber(void (*tsig) (int), void (*csig) (int), void (*asig) (int));



//__________
// session.c


void SessionWeg(long a);
void Zombie(long sid);
void BefehleWeg(long sid);
long SucheID(char *s);
short BefehlEinfuegen(short flag, short shift);
short SessionStatus(long sid);
short NeueTid(void);
void AufrufTriggern(void);

long FreierSpeicher(void);
void Statusbericht(void);




//_________
// zusatz.c


short der_select_test(void);
short punktpunkt_test(char *s);
long rechte_test(char *q, short dir);

void warten(long msec);



//_________
// client.c

void client(int socke, short neusessid);
void Reden(void);



//________
// tanne.c

short NeueSession(void);
short AufrufStarten(signed short sess_ID);

