
/* Copyright 2003 Uli Funcke & Klaus Kurzmann, FluxNetz GmbH */

/*
 * This file is part of Tanne
 *
 * Tanne is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 *
 * Tanne is distributed in the hope that it will be usefull,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tanne; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include "defines.h"



static struct sembuf p[1] = { {0, -1, 0} };
static struct sembuf v[1] = { {0, 1, 0} };
static struct sembuf pnw[1] = { {0, -1, IPC_NOWAIT} };






// ein neuer bereich wird erstellt: shm_id wird gltig
int shmStart(void)
{
	long f;
	long reserv;

	reserv = ANZ_IDS * sizeof(SBefehl) + ANZ_SESSIONS * sizeof(SSession);
#ifdef DEBUG2
	logger("shmStart(%d Bytes)", reserv);
#endif
	shm_id = shmget(IPC_PRIVATE, reserv, 0600 | IPC_CREAT);
	sem_id = semget(IPC_PRIVATE, 1, 0600 | IPC_CREAT);
	semctl(sem_id, 0, SETVAL, 1);

	if (shm_id == -1 || sem_id == -1)
		return (NEIN);

	if (!shmAn())
		return (NEIN);
	P();
	for (f = 0; f < ANZ_SESSIONS; f++)
		Session[f].status = FREI;
	for (f = 0; f < ANZ_IDS; f++)
		Befehl[f].sess_id = -1;
	V();
	shmAus();

#ifdef DEBUG2
	logger("shmStart() ging");
#endif
	return (JA);
}






// bereich wir freigegeben
void shmEnde(void)
{
	shmctl(shm_id, IPC_RMID, 0);
	semctl(sem_id, 0, IPC_RMID, 0);
#ifdef DEBUG2
	logger("shmEnde()");
#endif
}




//___________________________________________________________________________________________________________________
// 





// Ein Prozess ffnet bereich
int shmAn(void)
{
	char *s;

	s = (char *)shmat(shm_id, 0, 0);
	if (!s)
	{
		logger("FATAL: shmAn() failed");
		return (NEIN);
	}

	Session = (SSession *) s;
	Befehl = (SBefehl *) (s + ANZ_SESSIONS * sizeof(SSession));
#ifdef DEBUG3
	logger("shmAn() %d + %d * %d = %d", Session, ANZ_SESSIONS, sizeof(SSession), Befehl);
#endif
	return (JA);
}







// bereich schliessen

void shmAus(void)
{
	if (Session)
		shmdt(Session);
	Session = 0;
	Befehl = 0;
#ifdef DEBUG3
	logger("shmAus()");
#endif
}




//___________________________________________________________________________________________________________________
// 




// Semaphore P-Operation
// Eintritt/Warten

void P(void)
{
	int ret;

#ifdef DEBUG3
	logger("P()");
#endif
	sigprocmask(SIG_BLOCK, &signale, &sig_alt);
	ret = semop(sem_id, pnw, 1);
	if (ret == -1 && errno == EAGAIN)
		semop(sem_id, p, 1);				/* normales warten auf freigabe */
#ifdef DEBUG3
	logger("P() durch");
#endif
}






// Semaphore V-Operation
// Austritt/Freigabe

void V(void)
{
#ifdef DEBUG3
	logger("V()");
#endif
	semop(sem_id, v, 1);
	if (!sem_in_signal_flag)
		sigprocmask(SIG_UNBLOCK, &signale, &sig_alt);
#ifdef DEBUG3
	logger("V() durch");
#endif
}
