
/* Copyright 2003 Uli Funcke & Klaus Kurzmann, FluxNetz GmbH */

/*
 * This file is part of Tanne
 *
 * Tanne is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 *
 * Tanne is distributed in the hope that it will be usefull,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tanne; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include "defines.h"






void logger(char *str, ...)
{
#ifdef DEBUG_IN_FILE
	int h;
	time_t tim;
	struct tm *ptime;
#endif
	va_list args;
	char txt[512];

#ifdef DEBUG_IN_FILE
	time(&tim);
	ptime = localtime(&tim);
	snprintf(txt, 511, "%04d-%02d-%02d %02d:%02d:%02d (%d) ", ptime->tm_year + 1900, ptime->tm_mon + 1, ptime->tm_mday, ptime->tm_hour, ptime->tm_min, ptime->tm_sec, getpid());
#endif


#ifdef DEBUG_IN_FILE
	umask(0007);
	h = open(DEBUG_IN_FILE, O_WRONLY | O_CREAT | O_APPEND, 0660);
	if (h >= 0)
	{
		write(h, txt, strlen(txt));
		va_start(args, str);
		vsnprintf(txt, 511, str, args);
		va_end(args);
		write(h, txt, strlen(txt));
		write(h, "\n", 1);
		close(h);
	}
	else
	{
		va_start(args, str);
		vsnprintf(txt, 511, str, args);
		va_end(args);
		openlog("Tanne2", LOG_PID, LOG_DAEMON);
		syslog(LOG_INFO, "%s", txt);
		closelog();
	}
	umask(NORMALE_UMASK);
#else
	va_start(args, str);
	vsnprintf(txt, 511, str, args);
	va_end(args);
	openlog("Tanne2", LOG_PID, LOG_DAEMON);
	syslog(LOG_INFO, "%s", txt);
	closelog();
#endif
}






void ende(short sta)
{
	long f;
	static char *logsta[] = { "FREI", "RES", "STANDBY", "AKTIV", "ZOMBIE", "TRIG", "FATAL" };

	if (Session && sess_id >= 0)
	{
		P();
		if (Session[sess_id].status > FREI)
		{
			if (sta == ZOMBIE)
				BefehleWeg(sess_id);
			if (Session[sess_id].status != ZOMBIE)
			{
				Session[sess_id].status = sta;
#ifdef DEBUG
				logger("ende(%s) -> sess_id=%d", logsta[sta], sess_id);
#endif
			}
#ifdef DEBUG
			else
			{
				logger("ende(%s) -> sess_id=%d, ist aber schon ZOMBIE", logsta[sta], sess_id);
			}
#endif
		}
		else
			logger("OOOPS: ende(%s) free session %d terminated", logsta[sta], sess_id);
		V();
	}
#ifdef DEBUG
	else
		logger("ende(%s) aber kein Session-Array, sess_id=%d", logsta[sta], sess_id);
#endif

	for (f = 0; f < 256; f++)
		close(f);
	shmAus();
	exit(1);
}







int netz_init(void)
{
	socklen_t s;

	Gsocke = socket(AF_INET, SOCK_STREAM, 0);
	if (Gsocke == -1)
	{
		logger("socket() failed");
		return (NEIN);
	}

	in_name.sin_family = AF_INET;
#ifdef DEBUG
	in_name.sin_addr.s_addr = INADDR_ANY;
#else
	inet_aton("127.0.0.1", &in_name.sin_addr);
#endif
	in_name.sin_port = htons(TANNE_PORT);


	if (bind(Gsocke, (struct sockaddr *) &in_name, s = sizeof(in_name)) == -1)
	{
		logger("bind() failed");
		return (NEIN);
	}

	if (getsockname(Gsocke, (struct sockaddr *) &in_name, &s))
	{
		logger("getsockname() failed");
		return (NEIN);
	}

	listen(Gsocke, 10);
	return (JA);
}






void socken_init(long so)
{
	int s;

/*
	if( fcntl( Gsocke, F_SETFL, O_NONBLOCK ) < 0 )
		logger( "NONBLOCKING konnte nicht gesetzt werden" );
	
	if( fcntl( Gsocke, F_SETFL, O_NDELAY ) < 0 )
		logger( "NDELAY konnte nicht gesetzt werden" );
*/

	s = 1;
	if (setsockopt(so, IPPROTO_TCP, TCP_NODELAY, &s, sizeof(s)) != 0)
		logger("switching TCP_NODELAY on failed");
}





//void io_bereitung( long c )
//{
//  struct termios t;

//  if( fcntl( c, F_SETFL, O_SYNC ) < 0 )
//    logger( "O_SYNC konnte nicht gesetzt auf %d werden", c );

/*
	if( tcgetattr( c, &t ) < 0 )
		logger( "tcgetattr() failed" );
	t.c_lflag &= ~( ECHO | ICANON | IEXTEN | ISIG );
	t.c_iflag &= ~( BRKINT | ICRNL | INPCK | ISTRIP | IXON );
	t.c_cflag &= ~( CSIZE | PARENB );
	t.c_cflag |= CS8;
	t.c_oflag &= ~( OPOST );
	t.c_cc[VMIN] = 1;
	t.c_cc[VTIME] = 0;
	if( tcsetattr( c, TCSAFLUSH, &t ) < 0 )
		logger( "tcsetattr(%d) failed", c );
*/
//  if( ioctl( c, TIOCNOTTY, NEIN ) < 0 )
//    logger( "TIOCNOTTY failed at %d", c );
//}






#ifdef EIN_BYTE_LESEN
// Eine Zeile aus hand lesen und in zeile speichern
short lesen(long hand)
{
	char *s;
	long l;
	short anz_null;

	s = zeile;
	anz_null = 0;

#ifdef DEBUG2
	logger("lese aus %d", hand);
#endif
	do
	{
		if (s - zeile >= MAX_ZEILE)
		{
			logger("FATAL: LINE TO LONG!!!");
			ende(STANDBY);						// Zu lange Zeilen verschicken ziemt sich nicht
		}
		while ((l = read(hand, s, 1)) < 1)
		{
			if (l < 0)
			{
#ifdef DEBUG
				logger("Lesefehler in %d", hand);
#endif
				return (NEIN);
			}
			else
			{
				if (anz_null++ >= 2)
				{
					*s = 0;
					l = strlen(zeile);
#ifdef DEBUG
					if (l > 0)
						logger("FATAL: 4mal Null Bytes aus %d gelesen, '%s' (len = %d) umsonst", hand, zeile, l);
#endif
					return (NEIN);
				}
			}
		}
	}
	while (*s++ != 10);

	*(--s) = 0;										// terminieren

#ifdef DEBUG2
	logger("gelesen: '%s'", zeile);
#endif
	return (JA);
}





#else

// Eine Zeile aus hand lesen und in zeile speichern
// gepufferte Version

short lesen(long hand)
{
	long f;
	short i;
	short r;
	short anz_null;

	anz_null = 0;
	if (hand == Gsocke)
		i = 0;
	else
		i = 1;
#ifdef DEBUG2
	logger("lesen() i = %d", i);
#endif

	for (f = 0; f < MAX_ZEILE - 1; f++)
	{
		r = zeichenlesen(i, zeile + f);
		if (r == 0)
		{
			zeile[f] = 0;
#ifdef DEBUG
			logger("Lesefehler in %d", hand);
#endif
			return (NEIN);
		}
		if (r == 1)
		{
			if (anz_null++ >= 2)
			{
				zeile[f] = 0;
#ifdef DEBUG
				logger("Null Bytes aus %d gelesen", hand);
#endif
				return (NEIN);
			}
		}

		if (zeile[f] == '\n')
		{
			zeile[f] = 0;
#ifdef DEBUG2
			logger("gelesen: '%s'", zeile);
#endif
			return (JA);
		}
	}

	zeile[f] = 0;
#ifdef DEBUG2
	logger("gelesen: '%s'", zeile);
#endif
	return (JA);
}





long bytes_lesen(long hand, long anz)
{
	long f;
	short i;

	if (hand == Gsocke)
		i = 0;
	else
		i = 1;
	if (anz > MAX_ZEILE)
		anz = MAX_ZEILE;
	for (f = 0; f < anz; f++)
	{
		if (zeichenlesen(i, zeile + f) != 2)
			return (f);
	}

	return (anz);
}





short zeichenlesen(short i, char *z)
{
	long h;

	if (i == 0)
	{
		h = Gsocke;
	}
	else
	{
#ifdef USE_SOCKETPAIR
		h = GHand;
#else
		h = GHandL;
#endif
	}

	if (lesezeiger[i] == anz_gelesen[i])
	{
		lesezeiger[i] = 0;
#ifdef DEBUG2
		logger("read i = %d (d.h. hand = %d, i=0 heisst Socket, i=1 heisst pipe)", i, h);
#endif
		anz_gelesen[i] = read(h, lesebereich[i], MAX_LESEBEREICH);
#ifdef DEBUG2
		logger("%d Bytes bekommen", anz_gelesen[i]);
#endif

		if (anz_gelesen[i] == 0)
		{
			anz_gelesen[i] = lesezeiger[i] = 0;
			return (1);
		}

		if (anz_gelesen[i] < 0)
		{
			anz_gelesen[i] = lesezeiger[i] = 0;
			return (0);
		}
	}

	*z = lesebereich[i][lesezeiger[i]++];
	return (2);
}





#endif







// Sendet (iec) txt in hand
void senden(char *txt, long hand)
{
	long l, ol;
	char *s;

	ol = strlen(txt);
	s = txt;

#ifdef DEBUG2
	logger("sende: \"%s\"", txt);
#endif

	while (ol > 0)
	{
		l = write(hand, s, ol);
		if (l < 0)
		{
#ifdef DEBUG
			logger("Schreibfehler in %d", hand);
			logger("Sys_errlist: %s", sys_errlist[errno]);
#endif
			return;
		}

		if (!l)
			warten(800);
		ol -= l;
		s += l;
	}

	do
	{
		l = write(hand, "\n", 1);
		if (l < 0)
		{
#ifdef DEBUG
			logger("Schreibfehler auf %d", hand);
			logger("Sys_errlist: %s", sys_errlist[errno]);
#endif
			return;
		}

		if (!l)
			warten(200);
	}
	while (!l);
}







#ifdef EIN_BYTE_LESEN
// Durchreichemodus: einen Byte-Block lesen

long bytes_lesen(long hand, long anz)
{
	long l;

	if (anz > MAX_ZEILE)
		anz = MAX_ZEILE;
	l = read(hand, zeile, anz);
	if (l < 0)
	{
#ifdef DEBUG
		logger("lesefehler in %d", hand);
		logger("Sys_errlist: %s", sys_errlist[errno]);
#endif
		return (0);
	}

#ifdef DEBUG
	logger("%d Bytes gelesen", l);
#endif
	return (l);
}

#endif






// Sendet (iec) aus zeile in hand, ol Bytes
// Bytemodus zum durchreichen

void bytes_senden(long ol, long hand)
{
	long l;
	char *s;

	s = zeile;

	while (ol > 0)
	{
		l = write(hand, s, ol);

#ifdef DEBUG
		logger("%d Bytes gesendet", l);
#endif
		if (l < 0)
		{
#ifdef DEBUG
			logger("Schreibfehler in %d", hand);
			logger("Sys_errlist: %s", sys_errlist[errno]);
#endif
			return;
		}

		if (!l)
			warten(300);
		ol -= l;
		s += l;
	}
}
