
/*
    Copyright (c) 2000 FluxNetz GmbH

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "defines.h"





// in s Steht eine ID, die gesucht werden muss
// Rckgabe: BlockNr oder -1 (fr nicht vorh.)

long SucheID(char *s)
{
	long f;

#ifdef  DEBUG2
	logger("SucheID(%s)", s);
#endif

	P();
	for (f = 0; f < ANZ_IDS; f++)
	{
		if (Befehl[f].sess_id != -1 && strcmp(Befehl[f].id, s) == 0)
		{
#ifdef  DEBUG2
			logger("gefunden bei SessionID = %d -> '%s'", f, Befehl[f].bef);
#endif
			V();
			return (f);
		}
	}
	V();

#ifdef  DEBUG2
	logger("nichgefunden");
#endif
	return (-1);
}








short SessionStatus(long sid)
{
	short st;

	P();
	st = Session[sid].status;
	V();
	return (st);
}








// Alle Blocks mit sess_id weg

void SessionWeg(long sid)
{
	if (Session[sid].pid != -1)
	{
		logger("OOOPS: Could not clean session %ld, pid %ld is running", sid, Session[sid].pid);
		return;
	}

	BefehleWeg(sid);
#ifdef USE_SOCKETPAIR
	if (Session[sid].hand[0] != -1)
	{
#ifdef DEBUG
		logger("SessionWeg(%d) close(%d,%d)", sid, Session[sid].hand[0], Session[sid].hand[1]);
#endif
		close(Session[sid].hand[0]);
		close(Session[sid].hand[1]);
		Session[sid].hand[0] = -1;
	}
#else
	if (Session[sid].hl[0] != -1)
	{
		close(Session[sid].hl[0]);
		close(Session[sid].hl[1]);
		close(Session[sid].hs[0]);
		close(Session[sid].hs[1]);
		Session[sid].hl[0] = -1;
	}
#endif

	if (Session[sid].a_pid != -1)
	{
		logger("OOOPS: Could not clean session %ld, pid %ld is running", sid, Session[sid].a_pid);
		return;
	}

	Session[sid].status = FREI;
#ifdef DEBUG
	logger("Session %d ist wieder frei", sid);
#endif
}







void Zombie(long sid)
{
	BefehleWeg(sid);
	Session[sid].status = ZOMBIE;
#ifdef  DEBUG
	logger("Zombie(%d)", sid);
#endif
}







void BefehleWeg(long sid)
{
	long f, a;

	for (f = 0, a = 0; f < ANZ_IDS; f++)
	{
		if (Befehl[f].sess_id == sid)
		{
			Befehl[f].sess_id = -1;
			a++;
		}
	}
#ifdef  DEBUG
	logger("BefehleWeg(%d)  %d Befehle entfernt", sid, a);
#endif
}






long FreierSpeicher(void)
{
	long f, a;

	P();
	for (f = 0, a = 0; f < ANZ_IDS; f++)
	{
		if (Befehl[f].sess_id == -1)
			a += sizeof(SBefehl);
	}
	for (f = 0; f < ANZ_SESSIONS; f++)
	{
		if (Session[f].status == FREI)
			a += sizeof(SSession);
	}
	V();

	return (a);
}






// diese Funktion fgt den Befehl aus zeile in die Liste (bereich) ein
// erzeugt eine neue eindeutige ID und speichert diese (inclusive |) in
// zeile. Die Variable sess_id ist entweder noch auf 0 (Erster Befehl in diesem Aufruf)
// oder sie hat schon einen gltige Nummer erhalten.

short BefehlEinfuegen(short flag, short shift)
{
	long f;
	long m;

#ifdef  DEBUG2
	logger("BefehlEinfuegen(%s)", zeile);
#endif
	P();

	// leere id finden
	for (f = 0; f < ANZ_IDS; f++)
	{
		if (Befehl[f].sess_id == -1)	// gefunden
			break;										// aus for
	}

	if (f == ANZ_IDS)
	{
		V();
		logger("FATAL: out of memory - to many ID's");
		return (NEIN);
	}

	// f ist jetzt der richtige neue Bereich
	if (strlen(zeile) >= MAX_BEFEHL)
		logger("FATAL: command to long, truncated");
	strncpy(Befehl[f].bef, zeile + shift, MAX_BEFEHL);
	Befehl[f].bef[MAX_BEFEHL] = 0;
	Befehl[f].sess_id = sess_id;
	Befehl[f].flag = flag;
	V();
	// nun muss ein neuer ID_LAENGE langer Zufalls-String erzeugt werden

	m = 0;
	do
	{
		if (!NeueTid())
			return (NEIN);
		if (++m == 10)
		{
			logger("FATAL: 10 times get the same new TID. This is impossible!");
			return (NEIN);
		}
	}
	while (SucheID(zeile + 1) != -1);

	P();
	strcpy(Befehl[f].id, zeile + 1);
	V();
#ifdef  DEBUG2
	logger("BefehlEinfuegen() durch: '%s'", zeile);
#endif
	return (JA);
}







short NeueTid(void)
{
	long m, g;

	m = open("/dev/urandom", O_RDONLY | O_NONBLOCK);
	if (m < 0)
	{
		logger("FATAL, could not open /dev/urandom");
		return (NEIN);
	}

	zeile[0] = SPZ;
	for (g = 1; g < ID_LAENGE - 1;)
	{
		read(m, zeile + g, 1);
		if (zeile[g] >= '0' && zeile[g] <= '9')
			g++;
//    if( zeile[g] >= 'a' && zeile[g] <= 'z' )
//      g++;
		if (zeile[g] >= 'A' && zeile[g] <= 'Z')
			g++;
	}

	close(m);
	zeile[g] = 0;
	return (JA);
}









// in zeile[] steht ein Aufruf "@verz/prg"
// sess_id ist gesetzt auf 'ne neue id
// der Aufruf mu an den Vaterprozess weitergegeben werden
// und per sigalrm getriggert werden
// es darf aber nur ein Aufruf auf einmal gettigt werden
void AufrufTriggern(void)
{
	long f;
	short schlafpause;

#ifdef DEBUG
	logger("AufrufTriggern( %s )", zeile);
#endif
	do
	{
		schlafpause = NEIN;
		P();
		for (f = 0; f < ANZ_SESSIONS; f++)
		{
			if (Session[f].status == AUFRUF_TRIGGER)
			{
				schlafpause = JA;
				break;
			}
		}
		if (!schlafpause)
			Session[sess_id].status = AUFRUF_TRIGGER;
		V();
		if (schlafpause)
		{
#ifdef DEBUG
			logger("schlafpause eingelegt");
#endif
			warten(300);
		}
	}
	while (schlafpause);
#ifdef USE_SOCKETPAIR
	bytes_senden(MAX_ZEILE, GHand);	// die ganze Zeile, da ich eine feste Anzahl bytes senden will
#else
	bytes_senden(MAX_ZEILE, GHandS);
#endif
	kill(getppid(), SIGALRM);
#ifdef DEBUG
	logger("Alarmsignal gesendet sess_id = %d", sess_id);
#endif
}







void Statusbericht(void)
{
	long b, a, at, z, y, id, f, frei;

// Anzahl
// b = BELEGT, a = AKTIV, z = ZOMBIE, y = STANDBY, id = TID's

	at = a = b = z = y = id = 0;
	frei = 0;
	P();

	if (!Session)
	{
		V();
		logger("FATAL: no Session Variable");
		return;
	}

	for (f = 0; f < ANZ_SESSIONS; f++)
	{
		switch (Session[f].status)
		{
			case FREI:
				frei += sizeof(SSession);
				break;
			case AKTIV:
				a++;
				break;
			case BELEGT:
				b++;
				break;
			case ZOMBIE:
				z++;
				break;
			case STANDBY:
				y++;
				break;
			case AUFRUF_TRIGGER:
				at++;
				break;
		}
	}
	for (f = 0; f < ANZ_IDS; f++)
	{
		if (Befehl[f].sess_id != -1)
			id++;
		else
			frei += sizeof(SBefehl);
	}
	V();

#ifndef DEBUG2
	if (BETRAG(id - alt_freispeicher) > 70)
#endif
	{
		alt_freispeicher = id;
		logger("STATUS: %d run %d stdby %d res. %d zombies %d call %d IDs, %d bytes free", a, y, b, z, at, id, frei);
	}
}

