

/* Copyright 2003 Uli Funcke & Klaus Kurzmann, FluxNetz GmbH */

/*
 * This file is part of Tanne
 *
 * Tanne is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 *
 * Tanne is distributed in the hope that it will be usefull,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tanne; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include "defines.h"



// Alle 60 Sekunden einmal aufrumen vom Server

/*

1. alle Zombies befreien
2. alle abgelaufenen Sessions killen und freigeben

*/


void klingel(int na)
{
	long f, h, pid;

#ifdef DEBUG
	logger("klingel()");
#endif
	if (!shmAn())
	{
		logger("FATAL: shmAn() failed in klingel()");
		return;
	}

	sem_in_signal_flag = JA;
	Statusbericht();

	P();
	for (f = 0; f < ANZ_SESSIONS; f++)
	{
		if (Session[f].status == AUFRUF_TRIGGER)
		{
#ifdef USE_SOCKETPAIR
			h = Session[f].hand[1];
#else
			h = Session[f].hs[0];
#endif
#ifdef DEBUG3
			logger("h = %d", h);
#endif
			GHand = h;
			bytes_lesen(GHand, MAX_ZEILE);	// in AufrufTriggern() wurden genau MAX_ZEILE ber diese pipe geschickt
			zeile[MAX_ZEILE - 1] = 0;
			V();
			h = AufrufStarten(f);
			if (!h)
				SessionWeg(f);
			P();
		}
		else
		{
			if (Session[f].status == ZOMBIE)
			{
#ifdef DEBUG
				logger("Zombie-Session %d entfernt", f);
#endif
				SessionWeg(f);
			}
			else
			{
				if (Session[f].status == AKTIV || Session[f].status == STANDBY)
				{
					Session[f].ablauf--;
					if (Session[f].ablauf == 0)
					{
#ifdef USE_SOCKETPAIR
						h = Session[f].hand[0];
#else
						h = Session[f].hs[1];
#endif
						pid = Session[f].pid;
						logger("HUUPS: session timeout for session %d, PID=%d", f, pid);
						write(h, ENDE_BEFEHL, strlen(ENDE_BEFEHL));
					}
					if (Session[f].ablauf == -8)	// 4 Minuten spter
					{
						pid = Session[f].pid;
						if (pid == getpid())
						{
							logger("FATAL: System error: found my pid %d in session %f", pid, f);
							Session[f].pid = -1;
							Session[f].a_pid = -1;
							SessionWeg(f);		// das wars dann wohl
						}
						else
						{
							logger("FATAL: terminating session %d (PID=%d)", f, pid);
							if (pid > 1)
								kill(pid, SIGTERM);
						}
					}
					if (Session[f].ablauf == -12)	// noch zwei Minuten spter
					{
						pid = Session[f].pid;
						if (pid == getpid())
						{
							logger("FATAL: System error: found my pid %d in session %f", pid, f);
						}
						else
						{
							logger("FATAL: killing session %d (PID=%d)", f, pid);
							if (pid > 1)
								kill(pid, SIGKILL);
						}
						Session[f].pid = -1;
						Session[f].a_pid = -1;
						SessionWeg(f);			// das wars dann wohl
					}
				}
			}
		}
	}

	V();

	sem_in_signal_flag = NEIN;
	shmAus();
#ifdef DEBUG
	logger("klingel() durch");
#endif
}


//________________________________________________________________________________________________________



void terminator_server(int nu)
{
	close(Gsocke);
	shmEnde();
	logger("Tanne terminated.");
	exit(0);
}








void terminator_client(int nu)
{
#ifdef DEBUG2
	logger("terminator_client()");
#endif
	ende(STANDBY);
}










void kind_ende(int nu)
{
	long p, f;

#ifdef DEBUG
	logger("kind_ende()");
#endif
	if (!shmAn())
	{
		logger("FATAL: shmAn() failed in kind_ende()");
		return;
	}

	sem_in_signal_flag = JA;
	P();
	while ((p = wait3(0, WNOHANG | WUNTRACED, 0)) > 0)
	{
#ifdef DEBUG
		logger("PID = %d beendet", p);
#endif
		for (f = 0; f < ANZ_SESSIONS; f++)
		{
			if (Session[f].pid == p)
			{
				Session[f].pid = -1;
				logger("----------------- session %d terminated normally ----- (pid=%d) --------------", f, p);
				Zombie(f);
			}
			if (Session[f].a_pid == p)
				Session[f].a_pid = -1;
		}
	}

	V();
	sem_in_signal_flag = NEIN;
	shmAus();
#ifdef DEBUG
	logger("kind_ende() durch");
#endif
}




//___________________________________________________________________________________________________________________
// 







void server_signale(void)
{
	signalgeber(terminator_server, kind_ende, klingel);
}







void client_signale(void)
{
	signalgeber(terminator_client, NEIN, NEIN);
}









void signalgeber(void (*tsig) (int), void (*csig) (int), void (*asig) (int))
{
	struct itimerval t;

	sigemptyset(&signale);
	sigaddset(&signale, SIGTERM);
	sigaddset(&signale, SIGPIPE);

	if (csig)
		sigaddset(&signale, SIGCHLD);
	if (asig)
		sigaddset(&signale, SIGALRM);

	// --- TERM Signal ---
	sigh_term.sa_handler = tsig;
	sigh_term.sa_mask = signale;
	sigh_term.sa_flags = SA_RESTART;
	sigh_term.sa_restorer = NEIN;

	if ((sigaction(SIGTERM, &sigh_term, NULL)) == -1)
		FATAL("sigterm handler failed.");

	// --- CHILD Signal ---
	if (csig)
		sigh_child.sa_handler = csig;
	else
		sigh_child.sa_handler = SIG_DFL;
	sigh_child.sa_mask = signale;
	sigh_child.sa_flags = SA_RESTART;
	sigh_child.sa_restorer = NEIN;

	if ((sigaction(SIGCLD, &sigh_child, NULL)) == -1)
		FATAL("sigchld handler failed.");


	// --- ALARM Signal ---
	if (asig)
		sigh_alarm.sa_handler = asig;
	else
		sigh_alarm.sa_handler = SIG_DFL;
	sigh_alarm.sa_mask = signale;
	sigh_alarm.sa_flags = SA_RESTART;
	sigh_alarm.sa_restorer = NEIN;

	if ((sigaction(SIGALRM, &sigh_alarm, NULL)) == -1)
		FATAL("sigalarm handler failed.");


	// --- PIPE Signal ---
	sigh_pipe.sa_handler = SIG_IGN;
	sigh_pipe.sa_mask = signale;
	sigh_pipe.sa_flags = SA_RESTART;
	sigh_pipe.sa_restorer = NEIN;

	if ((sigaction(SIGPIPE, &sigh_pipe, NULL)) == -1)
		FATAL("sigpipe handler failed.");


	if (asig)
	{
		// alle 60 sekunden SIGALRM
		t.it_interval.tv_usec = 0;
		t.it_interval.tv_sec = 30;
		t.it_value.tv_usec = 0;
		t.it_value.tv_sec = 4;

		if (setitimer(ITIMER_REAL, &t, NULL))
			logger("setitimer() failed");
	}
#ifdef DEBUG2
	logger("signalgeber() durch");
#endif
}

